package reborncore.common.misc.vecmath;

import net.minecraft.util.BlockPos;
import net.minecraft.util.EnumFacing;
import net.minecraftforge.fml.common.FMLCommonHandler;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;
import net.minecraft.block.Block;
import net.minecraft.block.material.Material;
import net.minecraft.client.Minecraft;
import net.minecraft.init.Blocks;
import net.minecraft.server.MinecraftServer;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.Vec3;
import net.minecraft.world.World;


import java.util.StringTokenizer;

public class Vecs3d {
    protected double x, y, z;
    protected World w = null;

    public Vecs3d(double x, double y, double z) {

        this.x = x;
        this.y = y;
        this.z = z;
    }

    public Vecs3d(double x, double y, double z, World w) {

        this(x, y, z);
        this.w = w;
    }

    public Vecs3d(TileEntity te) {

        this(te.func_174877_v().func_177958_n(), te.func_174877_v().func_177956_o(), te.func_174877_v().func_177952_p(), te.func_145831_w());
    }

    public Vecs3d(Vec3 vec) {

        this(vec.field_72450_a, vec.field_72448_b, vec.field_72449_c);
    }

    public Vecs3d(Vec3 vec, World w) {

        this(vec.field_72450_a, vec.field_72448_b, vec.field_72449_c);
        this.w = w;
    }

    public boolean hasWorld() {

        return w != null;
    }

    public Vecs3d add(double x, double y, double z) {

        this.x += x;
        this.y += y;
        this.z += z;
        return this;
    }

    public Vecs3d add(EnumFacing dir) {

        return add(dir.func_82601_c(), dir.func_96559_d(), dir.func_82599_e());
    }

    public Vecs3d add(Vecs3d vec) {

        return add(vec.x, vec.y, vec.z);
    }

    public Vecs3d sub(double x, double y, double z) {

        this.x -= x;
        this.y -= y;
        this.z -= z;
        return this;
    }

    public Vecs3d sub(EnumFacing dir) {

        return sub(dir.func_82601_c(), dir.func_96559_d(), dir.func_82599_e());
    }

    public Vecs3d sub(Vecs3d vec) {

        return sub(vec.x, vec.y, vec.z);
    }

    public Vecs3d mul(double x, double y, double z) {

        this.x *= x;
        this.y *= y;
        this.z *= z;
        return this;
    }

    public Vecs3d mul(double multiplier) {

        return mul(multiplier, multiplier, multiplier);
    }

    public Vecs3d mul(EnumFacing direction) {

        return mul(direction.func_82601_c(), direction.func_96559_d(), direction.func_82599_e());
    }

    public Vecs3d multiply(Vecs3d v) {

        return mul(v.getX(), v.getY(), v.getZ());
    }

    public Vecs3d div(double x, double y, double z) {

        this.x /= x;
        this.y /= y;
        this.z /= z;
        return this;
    }

    public Vecs3d div(double multiplier) {

        return div(multiplier, multiplier, multiplier);
    }

    public Vecs3d div(EnumFacing direction) {

        return div(direction.func_82601_c(), direction.func_96559_d(), direction.func_82599_e());
    }

    public double length() {

        return Math.sqrt(x * x + y * y + z * z);
    }

    public Vecs3d normalize() {

        Vecs3d v = clone();

        double len = length();

        if (len == 0)
            return v;

        v.x /= len;
        v.y /= len;
        v.z /= len;

        return v;
    }

    public Vecs3d abs() {

        return new Vecs3d(Math.abs(x), Math.abs(y), Math.abs(z));
    }

    public double dot(Vecs3d v) {

        return x * v.getX() + y * v.getY() + z * v.getZ();
    }

    public Vecs3d cross(Vecs3d v) {

        return new Vecs3d(y * v.getZ() - z * v.getY(), x * v.getZ() - z
                * v.getX(), x * v.getY() - y * v.getX());
    }

    public Vecs3d getRelative(double x, double y, double z) {

        return clone().add(x, y, z);
    }

    public Vecs3d getRelative(EnumFacing dir) {

        return getRelative(dir.func_82601_c(), dir.func_96559_d(), dir.func_82599_e());
    }

    public EnumFacing getDirectionTo(Vecs3d vec) {

        for (EnumFacing d : EnumFacing.field_82609_l)
            if (getBlockX() + d.func_82601_c() == vec.getBlockX()
                    && getBlockY() + d.func_96559_d() == vec.getBlockY()
                    && getBlockZ() + d.func_82599_e() == vec.getBlockZ())
                return d;
        return null;
    }

    public boolean isZero() {

        return x == 0 && y == 0 && z == 0;
    }

    @Override
    public Vecs3d clone() {

        return new Vecs3d(x, y, z, w);
    }

    public boolean hasTileEntity() {

        if (hasWorld()) {
            return w.func_175625_s(getBlockPos()) != null;
        }
        return false;
    }

    public BlockPos getBlockPos(){
        return new BlockPos(x, y, z);
    }

    public TileEntity getTileEntity() {

        if (hasTileEntity()) {
            return w.func_175625_s(getBlockPos());
        }
        return null;
    }

    public boolean isBlock(Block b) {

        return isBlock(b, false);
    }

    public boolean isBlock(Block b, boolean checkAir) {

        if (hasWorld()) {
            Block bl = w.func_180495_p(getBlockPos()).func_177230_c();

            if (b == null && bl == Blocks.field_150350_a)
                return true;
            if (b == null && checkAir && bl.func_149688_o() == Material.field_151579_a)
                return true;
            if (b == null && checkAir && bl.isAir(w, getBlockPos()))
                return true;

            return bl.getClass().isInstance(b);
        }
        return false;
    }

    public Block getBlock() {

        return getBlock(false);
    }

    public Block getBlock(boolean airIsNull) {

        if (hasWorld()) {
            if (airIsNull && isBlock(null, true))
                return null;
            return w.func_180495_p(getBlockPos()).func_177230_c();

        }
        return null;
    }

    public World getWorld() {

        return w;
    }

    public Vecs3d setWorld(World world) {

        w = world;

        return this;
    }

    public double getX() {

        return x;
    }

    public double getY() {

        return y;
    }

    public double getZ() {

        return z;
    }

    public int getBlockX() {

        return (int) Math.floor(x);
    }

    public int getBlockY() {

        return (int) Math.floor(y);
    }

    public int getBlockZ() {

        return (int) Math.floor(z);
    }

    public double distanceTo(Vecs3d vec) {

        return distanceTo(vec.x, vec.y, vec.z);
    }

    public double distanceTo(double x, double y, double z) {

        double dx = x - this.x;
        double dy = y - this.y;
        double dz = z - this.z;
        return dx * dx + dy * dy + dz * dz;
    }

    public void setX(double x) {

        this.x = x;
    }

    public void setY(double y) {

        this.y = y;
    }

    public void setZ(double z) {

        this.z = z;
    }

    @Override
    public boolean equals(Object obj) {

        if (obj instanceof Vecs3d) {
            Vecs3d vec = (Vecs3d) obj;
            return vec.w == w && vec.x == x && vec.y == y && vec.z == z;
        }
        return false;
    }

    @Override
    public int hashCode() {

        return new Double(x).hashCode() + new Double(y).hashCode() << 8 + new Double(
                z).hashCode() << 16;
    }

    public Vec3 toVec3() {

        return new Vec3(x, y, z);
    }

    @Override
    public String toString() {

        String s = "Vector3{";
        if (hasWorld())
            s += "w=" + w.field_73011_w.func_177502_q() + ";";
        s += "x=" + x + ";y=" + y + ";z=" + z + "}";
        return s;
    }

    public EnumFacing toForgeDirection() {

        if (z == 1)
            return EnumFacing.SOUTH;
        if (z == -1)
            return EnumFacing.NORTH;

        if (x == 1)
            return EnumFacing.EAST;
        if (x == -1)
            return EnumFacing.WEST;

        if (y == 1)
            return EnumFacing.UP;
        if (y == -1)
            return EnumFacing.DOWN;

        return null;
    }

    public static Vecs3d fromString(String s) {

        if (s.startsWith("Vector3{") && s.endsWith("}")) {
            World w = null;
            double x = 0, y = 0, z = 0;
            String s2 = s.substring(s.indexOf("{") + 1, s.lastIndexOf("}"));
            StringTokenizer st = new StringTokenizer(s2, ";");
            while (st.hasMoreTokens()) {
                String t = st.nextToken();

                if (t.toLowerCase().startsWith("w")) {
                    int world = Integer.parseInt(t.split("=")[1]);
                    if (FMLCommonHandler.instance().getEffectiveSide()
                            .isServer()) {
                        for (World wo : MinecraftServer.func_71276_C().field_71305_c) {
                            if (wo.field_73011_w.func_177502_q() == world) {
                                w = wo;
                                break;
                            }
                        }
                    } else {
                        w = getClientWorld(world);
                    }
                }

                if (t.toLowerCase().startsWith("x"))
                    x = Double.parseDouble(t.split("=")[1]);
                if (t.toLowerCase().startsWith("y"))
                    y = Double.parseDouble(t.split("=")[1]);
                if (t.toLowerCase().startsWith("z"))
                    z = Double.parseDouble(t.split("=")[1]);
            }

            if (w != null) {
                return new Vecs3d(x, y, z, w);
            } else {
                return new Vecs3d(x, y, z);
            }
        }
        return null;
    }

    @SideOnly(Side.CLIENT)
    private static World getClientWorld(int world) {

        if (Minecraft.func_71410_x().field_71441_e.field_73011_w.func_177502_q() != world)
            return null;
        return Minecraft.func_71410_x().field_71441_e;
    }

}